from random import randrange, choice
from enum import Enum, auto

class Direction(Enum):
    NORTH = auto()
    SOUTH = auto()
    EAST = auto()
    WEST = auto()

class Outcome(Enum):
    MISS = auto()
    HIT = auto()
    SUNK = auto()

class AI(Enum):
    CHEAT = auto()
    EASY = auto()
    NORMAL = auto()
    HARD = auto()
    INSANE = auto()

class BgRepeat(Enum):
    NONE=auto()
    MENU=auto()
    ONCE=auto()
    LOOP=auto()

class Ship:
    def __init__(self, x, y, direction, size, name=""):
        self.x = x
        self.y = y
        self.size = size
        self.health = [1] * size
        self.direction = direction
        self.name = name

        self.ymin = y
        self.ymax = y
        self.xmin = x
        self.xmax = x
        if self.direction == Direction.EAST:
            self.xmax = x + (size - 1)
        elif self.direction == Direction.WEST:
            self.xmin = x - (size - 1)
        elif self.direction == Direction.SOUTH:
            self.ymax = y + (size - 1)
        elif self.direction == Direction.NORTH:
            self.ymin = y - (size - 1)

    def _pos_shoot(self,x,y):
        """Return shot position or None"""
        pos = None
        if x >= self.xmin and x <= self.xmax and y >= self.ymin and y <= self.ymax:
            pos = y-self.ymin + x-self.xmin
            if self.direction == Direction.WEST or self.direction == Direction.NORTH:
                pos = (self.size -1) - pos
        return pos

    def coords_of(self, i):
        """Return position of index or None"""
        if i < 0 or i >= self.size or self.direction is None:
            return None

        if self.direction == Direction.WEST:
            return (self.x-i, self.y)
        elif self.direction == Direction.EAST:
            return (self.x+i, self.y)
        elif self.direction == Direction.NORTH:
            return (self.x, self.y-i)
        elif self.direction == Direction.SOUTH:
            return (self.x, self.y+i)

    def _damage_pos(self, pos):
        """Decrease health of position"""
        if self.health[pos] > 0:
            self.health[pos] -= 1

    def get_health(self):
        """Return ship health"""
        return sum(self.health)

    def shoot(self, x, y):
        """Return Outcome for ship after shooting at x,y"""
        pos = self._pos_shoot(x, y)
        if pos is not None:
            old = self.get_health()
            self._damage_pos(pos)
            new = self.get_health()
            if new < old:
                if new > 0:
                    return Outcome.HIT
                else:
                    return Outcome.SUNK
            else:
                return Outcome.MISS

    def is_overlap(self, ship):
        """Return wether ship overlaps with self or not"""
        for x in range(min(self.xmin, self.xmax)-1, max(self.xmin,self.xmax)+1+1):
            for y in range(min(self.ymin, self.ymax)-1, max(self.ymin, self.ymax)+1+1):
                if ship._pos_shoot(x, y) is not None:
                    return True
        return False


class PlayerBoard:
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.ships = []
        self.hits = []
        self.miss = []
        self.sunk = []

    def add_ship(self, x, y, direction, size, name=""):
        ship = Ship(x, y, direction, size, name)
        if ship.xmax >= self.width or ship.xmin < 0 or ship.ymax >= self.height or ship.ymin < 0:
            raise ValueError

        if any([ship.is_overlap(s) for s in self.ships]):
            raise ValueError

        self.ships.append(ship)

    def shoot(self, x, y):
        if x < 0 or x >= self.width or y < 0 or y >= self.height:
            raise ValueError
        for ship in self.ships:
            outcome = ship.shoot(x,y)
            if outcome == Outcome.SUNK:
                self.hits.append((x,y))
                self.sunk.append(ship)
                return outcome
            elif outcome == Outcome.HIT:
                self.hits.append((x,y))
                return outcome
        self.miss.append((x,y))
        return Outcome.MISS

class Round:
    class State:
        stage=None
        options=None

        class Stage(Enum):
            MENU = auto()
            OPTIONS = auto()
            CREDITS = auto()
            NEWGAME = auto()
            PLAYER_SHIPS = auto()
            CPU_SHIPS = auto()
            PLAYER_TURN = auto()
            CPU_TURN = auto()
            NOTIFICATION = auto()
            GAMEOVER = auto()
            QUIT = auto()

        def __init__(self, stage, options={}):
            self.stage=stage
            self.options=options

        def switch(self, state_type, options={}, notification=None):
            if notification is None:
                self.stage=state_type
                self.options.clear()
                self.options=options.copy()
            else:
                self.stage=self.Stage.NOTIFICATION
                self.options.clear()
                self.options=options.copy()
                self.options['next_stage']=state_type
                self.options['next_options']=options
                self.options['message']=notification

    def __init__(self, width, ui, tui=None):
        if width > 26:
            raise ValueError
        self.width = width
        self.quit = False
        self.ui = ui
        self.textui = tui
        self.state = self.State(self.State.Stage.MENU)
        self.show_credits = True
        self.ships = [ ("The Anger", 2), ("The Victory Saber", 3), ("The Victory Saber", 3), ("The Pirate Storm", 4), ("The Princess", 5) ]
        self.player = None
        self.cpu = None

        self.params = sync_config(None)
        try:
            self.params = sync_config(None)
        except:
            self.params = {}

        for entry,default in {
                "ai": AI.NORMAL,
                "notif_timeout": 2000,
                "music": BgRepeat.ONCE,
                "sfx": True,
                "animations": True,
                }.items():
            self.params[entry] = self.params.get(entry, default)


    def update(self):
        if self.state.stage == self.State.Stage.QUIT:
            self.quit = True
        elif self.state.stage in [self.State.Stage.MENU, self.State.Stage.OPTIONS, self.State.Stage.CREDITS]:
            next_stage = self.state.options.get('choice', None)
            if next_stage is not None:
                self.state.switch(next_stage)
                sync_config(self.params)
        elif self.state.stage == self.State.Stage.NEWGAME:
            self.cpu = PlayerBoard(self.width, self.width)
            self.player = PlayerBoard(self.width, self.width)
            self.state.switch(self.State.Stage.PLAYER_SHIPS)
        elif self.state.stage == self.State.Stage.GAMEOVER:
            if self.state.options.get('skip', False):
                if self.show_credits == True:
                    self.state.switch(self.State.Stage.CREDITS)
                    self.show_credits = False
                else:
                    self.state.switch(self.State.Stage.MENU)
        elif self.state.stage == self.State.Stage.NOTIFICATION:
            if self.state.options.get('skip', False):
                self.state.switch(self.state.options['next_stage'], self.state.options['next_options'])
        elif self.state.stage == self.State.Stage.PLAYER_SHIPS:
            if self.state.options.get('ready', False) and self.state.options.get('x', None) is not None and self.state.options.get('y', None) is not None and self.state.options.get('direction', None) is not None:
                (shipname, shiplen) = self.ships[len(self.player.ships)]
                try:
                    self.player.add_ship(self.state.options['x'], self.state.options['y'], self.state.options['direction'], shiplen, shipname)
                except ValueError:
                    pass
                finally:
                    self.state.options = {}
            elif self.state.options.get('auto', False):
                valid_pos = False
                (shipname, shiplen) = self.ships[len(self.player.ships)]
                while not valid_pos:
                    try:
                        x,y = randrange(0, self.cpu.width), randrange(0, self.cpu.height)
                        self.player.add_ship(x, y, choice([Direction.NORTH, Direction.SOUTH, Direction.WEST, Direction.EAST]), shiplen, shipname)
                        valid_pos = True
                    except ValueError:
                        pass
                self.state.options.clear()

            if len(self.player.ships) == len(self.ships):
                self.state.switch(self.State.Stage.CPU_SHIPS)

        elif self.state.stage == self.State.Stage.CPU_SHIPS:
            while len(self.cpu.ships) < len(self.ships):
                x = randrange(0, self.cpu.width)
                y = randrange(0, self.cpu.height)
                direction = choice([Direction.NORTH, Direction.SOUTH, Direction.WEST, Direction.EAST])

                try:
                    (shipname, shiplen) = self.ships[len(self.cpu.ships)]
                    self.cpu.add_ship(x, y, direction, shiplen, shipname)
                except ValueError:
                    pass
            first = choice([self.State.Stage.PLAYER_TURN,self.State.Stage.CPU_TURN])
            first = self.State.Stage.PLAYER_TURN
            if first == self.State.Stage.PLAYER_TURN:
                self.state.switch(first, options={ "message": "You start." })
            else:
                self.state.switch(first, notification="CPU starts.")

        elif self.state.stage == self.State.Stage.PLAYER_TURN:
            if self.state.options.get('auto', False):
                while (self.state.options.get('x', None) is None and self.state.options.get('y', None) is None) or (self.state.options['x'],self.state.options['y']) in self.cpu.miss or (self.state.options['x'],self.state.options['y']) in self.cpu.hits:
                    self.state.options['x'] = randrange(0, self.cpu.width)
                    self.state.options['y'] = randrange(0, self.cpu.height)

            if (self.state.options.get('ready', False) or self.state.options.get('auto', False))  and self.state.options.get('x', None) is not None and self.state.options.get('y', None) is not None:
                if (self.state.options['x'], self.state.options['y']) in self.cpu.hits or (self.state.options['x'], self.state.options['y']) in self.cpu.miss:
                    self.state.options['ready'] = False
                    self.state.options['message'] = "Already shot there."
                else:
                    try:
                        outcome = self.cpu.shoot(self.state.options['x'], self.state.options['y'])
                        if outcome == Outcome.HIT:
                            self.state.switch(self.State.Stage.CPU_TURN, notification="You hit CPU.")
                        elif outcome == Outcome.MISS:
                            self.state.switch(self.State.Stage.CPU_TURN, notification="You missed CPU.")
                        elif len(self.cpu.sunk) == len(self.ships):
                            self.state.switch(self.State.Stage.GAMEOVER, options={'victory': True},  notification="You sunk CPU.")
                        else:
                            self.state.switch(self.State.Stage.CPU_TURN, notification="You sunk CPU.")
                    except ValueError:
                        pass
        elif self.state.stage == self.State.Stage.CPU_TURN:
            def get_rand(player):
                x,y = None,None
                while (x is None and y is None) or (x,y) in player.miss or (x,y) in player.hits:
                    x = randrange(0, player.width)
                    y = randrange(0, player.height)
                return (x,y)

            def interesting_coords(player):
                l = []
                for ship in player.ships:
                    if ship.get_health() > 0 and ship.get_health() < ship.size:
                        for i in range(ship.size):
                            if ship.health[i] == 0:
                                l.append(ship.coords_of(i))
                if len(l) == 0:
                    return None
                l = sorted(l, key=lambda s: 10*s[0] + s[1])

                xmin, ymin = xmax, ymax = l[0]

                while (xmax + 1, ymax) in l:
                    xmax += 1

                if xmin == xmax:
                    while (xmax, ymax + 1) in l:
                        ymax +=1

                return (xmin, xmax, ymin, ymax)

            if self.params['ai'] == AI.EASY or self.params['ai'] == AI.CHEAT:
                x, y = get_rand(self.player)
            elif self.params['ai'] == AI.NORMAL or self.params['ai'] == AI.HARD or self.params['ai'] == AI.INSANE:
                i = interesting_coords(self.player)
                if i is None:
                    x, y = get_rand(self.player)
                else:
                    (xmin, xmax, ymin, ymax) = i
                    if xmin < xmax:
                        y = ymin
                        if xmin==0 or (xmin-1,y) in self.player.miss:
                            x = xmax+1
                        elif xmax==self.width-1 or (xmax+1, y) in self.player.miss:
                            x = xmin-1
                        else:
                            x = choice([xmin-1, xmax+1])
                    elif ymin < ymax:
                        x = xmin
                        if ymin==0 or (x,ymin-1) in self.player.miss:
                            y = ymax+1
                        elif ymax==self.width-1 or (x, ymax+1) in self.player.miss:
                            y = ymin-1
                        else:
                            y = choice([ymin-1, ymax+1])
                    else:
                        p = [(xmin-1, ymin), (xmin+1, ymin), (xmin, ymin-1), (xmin, ymin+1)]
                        l = []
                        for c in p:
                            if c not in self.player.miss and all([c[i]>=0 and c[i]<self.width for i in [0, 1]]):
                                l.append(c)
                        (x,y) = choice(l)

            outcome = self.player.shoot(x,y)

            if outcome == Outcome.HIT:
                self.state.switch(self.State.Stage.PLAYER_TURN, options={ "message": "CPU hit you."})
            elif outcome == Outcome.MISS:
                self.state.switch(self.State.Stage.PLAYER_TURN, options={ "message": "CPU missed you."})
            else:
                if len(self.player.sunk) == len(self.ships):
                    self.state.switch(self.State.Stage.GAMEOVER, options={ "message": "CPU sunk you.", "victory": False})
                else:
                    self.state.switch(self.State.Stage.PLAYER_TURN, options={ "message": "CPU sunk you."})

    def run(self):
        while not self.quit:
            self.update()
            self.ui.draw(self)
            self.textui.draw(self)
            self.ui.read_events(self)

def sync_config(data=None):
    from configparser import ConfigParser
    import os

    config = ConfigParser()

    xdgprefix = os.getenv('XDG_CONFIG_HOME', os.path.expanduser('~/.config'))
    dirname = os.path.join(xdgprefix, "capbattleship")
    os.makedirs(dirname, exist_ok=True)

    filename = os.path.join(dirname, "config")

    if data is not None:
        for (key, value) in data.items():
            config['DEFAULT'][key] = str(value)
        with open(filename, 'w') as configfile:
            config.write(configfile)
    elif os.path.isfile(filename):
        config.read(filename)
        data = {}
        for key in ["sfx", "animations"]:
            try:
                data[key] = config['DEFAULT'].getboolean(key)
            except:
                pass
        for key in ["notif_timeout"]:
            try:
                data[key] = config['DEFAULT'].getfloat(key)
            except:
                pass
        try:
            data['ai'] = AI.__members__[config['DEFAULT']['ai'][3:]]
        except:
            pass
        try:
            data['music'] = BgRepeat.__members__[config['DEFAULT']['music'][9:]]
        except:
            pass

        return data
    else:
        return {}
