import math
import gettext
import locale
from enum import Enum, auto

from .pygamesprites import *
from .logic import Direction, Ship, Round, AI, BgRepeat
Stage = Round.State.Stage

WHITE = (255, 255, 255)
ORANGE = (255, 127, 0)
BLACK = (0, 0, 0)
LIGHTGRAY = (0xd9, 0xd9, 0xd9)
DARKGRAY = (0x2f, 0x2f, 0x2f)
MEDIUMGRAY = (0x88, 0x88, 0x88)

FONT="fonts/DejaVuSans.ttf"

def ts(*l):
    return tuple([sum(t) for t in zip(*l)])

def cut_text(width, text):
    out = []
    while len(text) > width:
        i = text[:width].rindex(" ")
        out.append(text[:i])
        text = text[i+1:]
    out.append(text)
    return out


class GuiMenu:
    def __init__(self):
        splashscreen = pygame.image.load("gfx/splashscreen.jpg").convert()
        self.bg = splashscreen

        menu_font = pygame.font.Font(FONT, 20)
        self.menu_items = []
        self.menu_items.append((
            menu_font.render(_("Start the game"), 1, ORANGE),
            menu_font.render(_("Start the game"), 1, WHITE),
            'n',
            Stage.NEWGAME,
            ))
        self.menu_items.append((
            menu_font.render(_("Manage options"), 1, ORANGE),
            menu_font.render(_("Manage options"), 1, WHITE),
            'o',
            Stage.OPTIONS,
            ))
        self.menu_items.append((
            menu_font.render(_("See credits"), 1, ORANGE),
            menu_font.render(_("See credits"), 1, WHITE),
            'c',
            Stage.CREDITS,
            ))
        self.menu_items.append((
            menu_font.render(_("Quit"), 1, ORANGE),
            menu_font.render(_("Quit"), 1, WHITE),
            'q',
            Stage.QUIT,
            ))
        self.menu_topleft = (550, 600)

        self.menu_rects = []
        x,y = self.menu_topleft
        for (n,(active, inactive, key, stage)) in enumerate(self.menu_items):
            rect = active.get_rect()
            rect.topleft = (x,y)
            self.menu_rects.append(rect)
            y += rect.height + 10
            x += 20

        self.reset()

    def reset(self):
        self.menu_index = 0

    def draw(self, s):
        update_rects = []

        s.blit(self.bg, (0, 0))
        update_rects.append(s.get_rect())

        for (n, (active, inactive, key, stage)) in enumerate(self.menu_items):
            if self.menu_index == n:
                item = active
            else:
                item = inactive
            s.blit(item, self.menu_rects[n])
            update_rects.append(self.menu_rects[n])
        return update_rects

    def update(self, rnd, params):
        pass

    def input(self, event, rnd):
        if event.type == pygame.MOUSEBUTTONDOWN:
            for (n, (active, inactive, key, stage)) in enumerate(self.menu_items):
                if self.menu_rects[n].collidepoint(event.pos):
                    rnd.state.options['choice'] = stage
        elif event.type == pygame.KEYDOWN:
            for (active, inactive, k, stage) in self.menu_items:
                if event.unicode == k:
                    rnd.state.options['choice'] = stage
            if event.key == pygame.K_DOWN:
                self.menu_index += 1
                if self.menu_index == len(self.menu_items):
                    self.menu_index = 0
                rnd.state.options['needs_redraw'] = True
            elif event.key == pygame.K_UP:
                self.menu_index -= 1
                if self.menu_index == -1:
                    self.menu_index += len(self.menu_items)
                rnd.state.options['needs_redraw'] = True
            elif event.key == pygame.K_RETURN:
                rnd.state.options['choice'] = self.menu_items[self.menu_index][3]
        elif event.type == pygame.MOUSEMOTION:
            for i, r in enumerate(self.menu_rects):
                if r.collidepoint(event.pos):
                    self.menu_index = i
                    rnd.state.options['needs_redraw'] = True

class GuiOptions:
    def __init__(self):
        options  = pygame.image.load("gfx/options.jpg").convert()

        self.bg = options

        menu_font = pygame.font.Font(FONT, 20)
        self.menu_items = []
        self.menu_items.append((
            menu_font.render(_("Menu"), 1, ORANGE),
            menu_font.render(_("Menu"), 1, WHITE),
            'm',
            Stage.MENU,
            ))
        self.menu_center = (512, 700)

        self.menu_rects = []
        x,y = self.menu_center
        for (n,(active, inactive, key, stage)) in enumerate(self.menu_items):
            rect = active.get_rect()
            rect.center = (x,y)
            self.menu_rects.append(rect)
            y += rect.height + 10
            x += 20

        font_options = pygame.font.Font(FONT, 20)
        font_items = pygame.font.Font(FONT, 18)

        options_options = [
                (_("Difficulty"), [
                    (_("Cheat"),  AI.CHEAT),
                    (_("Easy"),   AI.EASY),
                    (_("Normal"), AI.NORMAL),
                    #(_("Hard"),   AI.HARD),
                    #(_("Insane"), AI.INSANE),
                    ],
                    "ai"),
                (_("Notification speed"), [
                    (_("Wait"),   float('inf')),
                    (_("2s"),     2000),
                    (_("1s"),     1000),
                    (_("0.5s"),    500),
                    (_("Skip"),      0),
                    ],
                    "notif_timeout"),
                (_("Background Music"), [
                    (_("Disabled"),  BgRepeat.NONE),
                    (_("Menu only"), BgRepeat.MENU),
                    (_("Play once"), BgRepeat.ONCE),
                    (_("Repeat"),    BgRepeat.LOOP),
                    ],
                    "music"),
                (_("Sfx"), [
                    (_("Disabled"), False),
                    (_("Enabled"),  True),
                    ],
                    "sfx"),
                (_("Animations"), [
                    (_("Disabled"), False),
                    (_("Enabled"),  True),
                    ],
                    "animations"),
                #(_("Fullscreen"), [
                #    (_("Disabled"), False),
                #    (_("Enabled"), True),
                #    ],
                #    "fullscreen"),
                ]

        self.options_config = {}
        self.options_menu = [{
            "active":   font_options.render(label, 1, ORANGE),
            "inactive": font_options.render(label, 1, WHITE),
            "option":   name,
            "items": [{
                "active":     font_options.render(ilabel, 1, ORANGE),
                "inactive":   font_options.render(ilabel, 1, WHITE),
                "active_u":   font_options.render(ilabel, 1, LIGHTGRAY),
                "inactive_u": font_options.render(ilabel, 1, MEDIUMGRAY),
                "value":      ivalue,
                }
                for (ilabel, ivalue) in items ],
            }
            for (label, items, name) in options_options ]


        options_topleft = (70, 150)

        x,y = options_topleft
        for option in self.options_menu:
            r = option['active'].get_rect()
            r.topleft = (x, y)
            option['rect'] = r
            (ix, iy) = (350, r.bottom)
            for item in option['items']:
                ir = item['active'].get_rect()
                ir.bottomleft = (ix + 30, iy)
                item['rect'] = ir
                (ix, iy) = ir.bottomright
            x,y = r.bottomleft
            y+=20

        self.reset()

    def reset(self):
        self.menu_index = 0
        self.options_active = 0
        self.options_config.clear()
        self.options_config_loaded = False

    def draw(self, s):
        update_rects = []

        s.blit(self.bg, (0, 0))
        update_rects.append(s.get_rect())

        for n, option in enumerate(self.options_menu):
            if n == self.options_active:
                s.blit(option['active'], option['rect'])
            else:
                s.blit(option['inactive'], option['rect'])
            for m, item in enumerate(option['items']):
                if n == self.options_active:
                    if self.options_config.get(n, None) == m:
                        s.blit(item['active'], item['rect'])
                    else:
                        s.blit(item['inactive'], item['rect'])
                else:
                    if self.options_config.get(n, None) == m:
                        s.blit(item['active_u'], item['rect'])
                    else:
                        s.blit(item['inactive_u'], item['rect'])

        for (n, (active, inactive, key, stage)) in enumerate(self.menu_items):
            if self.menu_index == n:
                item = active
            else:
                item = inactive
            s.blit(item, self.menu_rects[n])
            update_rects.append(self.menu_rects[n])
        return update_rects

    def update(self, rnd, params):
        # Load config options
        if not self.options_config_loaded:
            for n,option in enumerate(self.options_menu):
                for m,item in enumerate(option['items']):
                    if item['value'] == params.get(option['option'], None):
                        self.options_config[n] = m
                # fallback on first value
                if self.options_config.get(n, None) is None:
                    self.options_config[n] = self.options_menu[n]['items'][0]['value']
            self.options_config_loaded = True


        # Apply config options
        for (n, v) in self.options_config.items():
            name = self.options_menu[n]['option']
            params[name] = self.options_menu[n]['items'][v]['value']

    def input(self, event, rnd):
        if event.type == pygame.MOUSEBUTTONDOWN:
            for (n, (active, inactive, key, stage)) in enumerate(self.menu_items):
                if self.menu_rects[n].collidepoint(event.pos):
                    rnd.state.options['choice'] = stage
            for n, option in enumerate(self.options_menu):
                for m, item in enumerate(option['items']):
                    if item['rect'].collidepoint(event.pos):
                        self.options_config[n] = m
                        rnd.state.options['needs_redraw'] = True
                        return
        elif event.type == pygame.KEYDOWN:
            for (active, inactive, k, stage) in self.menu_items:
                if event.unicode == k:
                    rnd.state.options['choice'] = stage
            if event.key == pygame.K_RETURN:
                rnd.state.options['choice'] = self.menu_items[self.menu_index][3]
            elif event.key in [pygame.K_DOWN, pygame.K_UP, pygame.K_LEFT, pygame.K_RIGHT] and self.options_active is not None and self.options_config.get(self.options_active, None) is not None:
                rnd.state.options['needs_redraw'] = True
                if self.options_active is None:
                    self.options_active = 0
                elif event.key == pygame.K_DOWN:
                    self.options_active += 1
                    self.options_active %= len(self.options_menu)
                elif event.key == pygame.K_UP:
                    self.options_active += len(self.options_menu) - 1
                    self.options_active %= len(self.options_menu)
                elif event.key == pygame.K_LEFT:
                    self.options_config[self.options_active] += len(self.options_menu[self.options_active]['items']) - 1
                    self.options_config[self.options_active] %= len(self.options_menu[self.options_active]['items'])
                elif event.key == pygame.K_RIGHT:
                    self.options_config[self.options_active] += 1
                    self.options_config[self.options_active] %= len(self.options_menu[self.options_active]['items'])
        elif event.type == pygame.MOUSEMOTION:
            for i, r in enumerate(self.menu_rects):
                if r.collidepoint(event.pos):
                    self.menu_index = i
                    rnd.state.options['needs_redraw'] = True
            for n, option in enumerate(self.options_menu):
                if option['rect'].collidepoint(event.pos):
                    self.options_active = n
                    rnd.state.options['needs_redraw'] = True
                    return
                for m, item in enumerate(option['items']):
                    if item['rect'].collidepoint(event.pos):
                        self.options_active = n
                        rnd.state.options['needs_redraw'] = True
                        return

class GuiCredits:
    def __init__(self):
        credits  = pygame.image.load("gfx/credits.jpg").convert()

        self.quotes = [
                {
                    "author": "Fabien Givors (captnfab)",
                    "descr": _("Source code development (uses python3, pygame), translation (fr, en)"),
                    "url": "https://chezlefab.net",
                    "licence": "GPL + MIT",
                    },
                {
                    "author": "Damien Monteillard (lastrodamo)",
                    "descr": _("Graphism, game design (illustrations, ships, VFX, Map)"),
                    "url": "https://www.3dminfographie.com/",
                    "licence": "CC0",
                    },
                {
                    "author": "seeker47",
                    "descr": _("100 Golden Hern piece (Coins)"),
                    "url": "https://blendswap.com/blend/17578",
                    "licence": "CC0",
                    },
                {
                    "author": "'Kyd' (Joseph Clayton Clarke)",
                    "descr": _("Mr. Bumble (Inspiration for 3D model)"),
                    "url": "https://en.wikipedia.org/wiki/Bumble_(Oliver_Twist)",
                    },
                {
                    "author": "yd",
                    "descr": _("Container Pack (Barrels)"),
                    "url": "https://blendswap.com/blend/16194",
                    "licence": "CC0",
                    },
                {
                    "author": "Scott Buckley",
                    "descr": _("Titan (Music)"),
                    "url": "https://www.scottbuckley.com.au/",
                    "licence": "CC-BY 4.0",
                    },
                {
                    "author": "reklamacja",
                    "descr": _("Stone in the water 080117-002 drop (sound effect)"),
                    "url": "https://freesound.org/people/reklamacja/sounds/342457/",
                    "licence": "CC0",
                    },
                {
                    "author": "hellboy1305",
                    "descr": _("Canon Gun - Canon sound_01 (sound effect)"),
                    "url": "https://freesound.org/people/hellboy1305/sounds/370273/",
                    "licence": "CC0",
                    },
                {
                    "author": _("All the testers"),
                    "descr": _("Thanks! (testing, advices, bugs)"),
                    }
                ]

        font_descr = pygame.font.Font(FONT, 18)
        font_name = pygame.font.Font(FONT, 16)
        font_name.set_bold(False)
        font_url = pygame.font.Font(FONT, 16)
        font_url.set_italic(True)
        font_licence = pygame.font.Font(FONT, 16)

        topleft = (50, 70)
        for q in self.quotes:
            descr = font_descr.render(q['descr'], 1, WHITE)
            descr_r = descr.get_rect()
            descr_r.topleft = topleft

            name = font_name.render(q['author'], 1, ORANGE)
            name_r = name.get_rect()
            (x, y) = descr_r.bottomleft
            name_r.topleft = (x, y+5)

            url = font_name.render(q.get('url',' '), 1, WHITE)
            url_r = url.get_rect()
            (x, y) = name_r.bottomright
            url_r.bottomleft = (x+15, y)

            licence = font_licence.render(q.get('licence', ' '), 1, ORANGE)
            licence_r = licence.get_rect()
            (x, y) = url_r.bottomright
            licence_r.bottomleft = (x+15, y)

            credits.blit(descr, descr_r)
            credits.blit(name,  name_r)
            credits.blit(url,  url_r)
            credits.blit(licence,  licence_r)

            (x, y) = name_r.bottomleft
            topleft = (x, y+25)

        self.bg = credits

        menu_font = pygame.font.Font(FONT, 20)
        self.menu_items = []
        self.menu_items.append((
            menu_font.render(_("Menu"), 1, ORANGE),
            menu_font.render(_("Menu"), 1, WHITE),
            'm',
            Stage.MENU,
            ))
        self.menu_center = (512, 700)

        self.menu_rects = []
        x,y = self.menu_center
        for (n,(active, inactive, key, stage)) in enumerate(self.menu_items):
            rect = active.get_rect()
            rect.center = (x,y)
            self.menu_rects.append(rect)
            y += rect.height + 10
            x += 20

        self.reset()

    def reset(self):
        self.menu_index = 0

    def draw(self, s):
        update_rects = []

        s.blit(self.bg, (0, 0))
        update_rects.append(s.get_rect())

        for (n, (active, inactive, key, stage)) in enumerate(self.menu_items):
            if self.menu_index == n:
                item = active
            else:
                item = inactive
            s.blit(item, self.menu_rects[n])
            update_rects.append(self.menu_rects[n])
        return update_rects

    def update(self, rnd, params):
        pass

    def input(self, event, rnd):
        if event.type == pygame.MOUSEBUTTONDOWN:
            for (n, (active, inactive, key, stage)) in enumerate(self.menu_items):
                if self.menu_rects[n].collidepoint(event.pos):
                    rnd.state.options['choice'] = stage
        elif event.type == pygame.KEYDOWN:
            for (active, inactive, k, stage) in self.menu_items:
                if event.unicode == k:
                    rnd.state.options['choice'] = stage
            if event.key == pygame.K_DOWN:
                self.menu_index += 1
                if self.menu_index == len(self.menu_items):
                    self.menu_index = 0
                rnd.state.options['needs_redraw'] = True
            elif event.key == pygame.K_UP:
                self.menu_index -= 1
                if self.menu_index == -1:
                    self.menu_index += len(self.menu_items)
                rnd.state.options['needs_redraw'] = True
            elif event.key == pygame.K_RETURN:
                rnd.state.options['choice'] = self.menu_items[self.menu_index][3]
        elif event.type == pygame.MOUSEMOTION:
            for i, r in enumerate(self.menu_rects):
                if r.collidepoint(event.pos):
                    self.menu_index = i
                    rnd.state.options['needs_redraw'] = True

class GuiSetupShips:
    def coords_to_pos(self, coords, topleft):
        y,x = coords
        a = self.cell_tr[0]
        b = self.cell_tl[0]
        c = self.cell_tr[1]
        d = -self.cell_tl[1]

        bx = x*a + y*b
        by = x*c + y*d

        by += 10*self.cell_tl[1]

        # ocean offset
        bx += topleft[0]
        by += topleft[1]
        return (int(bx), int(by))

    def pos_to_coords(self, pos, topleft):
        bx, by = pos
        # ocean offset
        bx -= topleft[0]
        by -= topleft[1]

        by -= 10*self.cell_tl[1]

        a = -self.cell_tl[1]
        b = -self.cell_tl[0]
        c = -self.cell_tr[1]
        d =  self.cell_tr[0]

        det = a*d-b*c
        x, y = (a*bx + b*by)/det, (c*bx + d*by)/det
        x, y = math.floor(x), math.floor(y)

        return (y,x)

    def __init__(self):
        self.reset()

        bg = pygame.Surface((1024, 768))

        bigboard_rect     = pygame.Rect((0, 0), (bg.get_width(), 614))
        hitboard_rect     = pygame.Rect((0, 614), (bg.get_width(), bg.get_height()-614))
        pygame.draw.rect(bg, DARKGRAY,  bigboard_rect)
        pygame.draw.rect(bg, LIGHTGRAY, hitboard_rect)

        font_player = pygame.font.Font(FONT, 16)
        label_player = font_player.render(_("Player"), 1, WHITE)
        bg.blit(label_player, (10, 10))

        self._prepare_grid(bg)

        pygame.draw.line(bg, MEDIUMGRAY, (10, hitboard_rect.bottom - 35), (10 + hitboard_rect.width/2 - 10, hitboard_rect.bottom - 35), 1)

        self.bg = bg

        help_pane = pygame.Surface((1024, 768)).convert_alpha()
        help_pane.fill((0, 0, 0, 0))
        self._prepare_help_pane(help_pane)
        self.help_pane = help_pane

        hlp_button_active   = pygame.image.load("gfx/help_button_focus.png").convert_alpha()
        hlp_button_inactive = pygame.image.load("gfx/help_button.png").convert_alpha()
        hlp_button_width    = 50
        #hlp_button_active   = pygame.transform.scale(hlp_button_active, (hlp_button_width, hlp_button_width))
        #hlp_button_inactive = pygame.transform.scale(hlp_button_inactive, (hlp_button_width, hlp_button_width))

        self.hlp_button_topright = (1014, 10)

        self.menu_items = []
        self.menu_items.append((
            hlp_button_active,
            hlp_button_inactive,
            'h',
            None
            ))

        self.menu_rects = []
        rect = hlp_button_active.get_rect()
        rect.topright = self.hlp_button_topright
        self.menu_rects.append(rect)

        self._prepare_ship_descr()

        self.notif_rect   = pygame.Rect((0, 450), (bg.get_width(), 50))

    def _prepare_help_pane(self, s):
        hlp_move =   pygame.image.load("gfx/help_move-boat.png").convert_alpha()
        hlp_rotate = pygame.image.load("gfx/help_rotate-boat.png").convert_alpha()
        hlp_ready =  pygame.image.load("gfx/help_ready-boat.png").convert_alpha()

        hlp_move_width = 200
        hlp_rotate_width = 200
        hlp_ready_width = 200
        hlp_move_height = int(hlp_move.get_height()*hlp_move_width/hlp_move.get_width())
        #hlp_move = pygame.transform.smoothscale(hlp_move, (hlp_move_width, hlp_move_height))
        hlp_rotate_height = int(hlp_rotate.get_height()*hlp_rotate_width/hlp_rotate.get_width())
        #hlp_rotate = pygame.transform.smoothscale(hlp_rotate, (hlp_rotate_width, hlp_rotate_height))
        hlp_ready_height = int(hlp_ready.get_height()*hlp_ready_width/hlp_ready.get_width())
        #hlp_ready = pygame.transform.smoothscale(hlp_ready, (hlp_ready_width, hlp_ready_height))

        help_top = 25
        help_left = 550

        s.blit(hlp_move,   (help_left, help_top))
        s.blit(hlp_rotate, (help_left + hlp_move.get_width(), help_top))
        s.blit(hlp_ready,  (help_left + hlp_move.get_width()/2,  help_top + hlp_move.get_height() + 50))


    def _prepare_ship_descr(self):
        shipname_font = pygame.font.Font(FONT, 36)
        shipdescr_font = pygame.font.Font(FONT, 14)
        descr_width = 55
        self.ships_descr = [None, None]
        for name, size, descr in [
                (_("The Anger"),         2, _("Every shipman is only grousing day and night, which drives the captain completely mad. There is one thing you can do to help him unwind, that is be an easy pray. I would advise you not to…")),
                (_("The Victory Saber"), 3, _("They were twin brothers, incredibly talentuous swashbucklers, and the fell for a mysterious lady. She said only the most fearsome pirate would be her match. Will you try their cruelty?")),
                (_("The Pirates Storm"), 4, _("This vessel is known for its sneaky attacks during storms. Manœuvrable and fast, no ship can escape her in the raging waters. If the weather is taking a turn for the worse, be sure that you're alone…")),
                (_("The Princess"),      5, _("Rather than the vessel, 'The Princess' refers to its heartless captain. Famous buccaneer, she is said to be as rich as a king. Well, if she meets you, maybe she well become even richer…")),
                ]:
            ship = {}
            ship['name']        = name
            ship['rname']       = shipname_font.render(name, 1, WHITE)
            ship['descr']       = descr
            ship['rdescr']      = [ shipdescr_font.render(line, 1, WHITE) for line in cut_text(descr_width, descr)]
            ship['turntable']   = ShipTurntable(ship_size=size, pos=(750, 200), loop=True, resize_width=400)
            models = {}
            for direction,letter in [(Direction.NORTH, "N"), (Direction.SOUTH, "S"), (Direction.EAST, "E"), (Direction.WEST, "W"), ("P", "P")]:
                view = pygame.image.load("gfx/boat_"+str(size)+"_"+letter+".png").convert_alpha()
                #view = pygame.transform.smoothscale(view, (int(5*self.cell_tl[0]),int(5*self.cell_tl[0])))
                models[direction] = view
            ship['models'] = models
            self.ships_descr.append(ship)

        self.ships_descr[2]['offset'] = {
                Direction.NORTH: ( 17,-17),
                Direction.SOUTH: ( 30, -7),
                Direction.EAST:  ( 36,-23),
                Direction.WEST:  ( 10,-14),
                'P': ( 0,  0),
                }
        self.ships_descr[3]['offset'] = {
                Direction.NORTH: ( 10,-25),
                Direction.SOUTH: ( 41,  4),
                Direction.EAST:  ( 47,-27),
                Direction.WEST:  ( -5, -5),
                'P': ( 0,  0),
                }
        self.ships_descr[4]['offset'] = {
                Direction.NORTH: ( -2,-36),
                Direction.SOUTH: ( 53, 16),
                Direction.EAST:  ( 69,-36),
                Direction.WEST:  (-16,  1),
                'P': ( 0,  0),
                }
        self.ships_descr[5]['offset'] = {
                Direction.NORTH: ( -9, -45),
                Direction.SOUTH: ( 66,  32),
                Direction.EAST:  ( 81, -44),
                Direction.WEST:  (-36,  11),
                'P': ( 0,  0),
                }

    def _draw_ship(self, s, ship, board_topleft):
        cx, cy = self.coords_to_pos((ship.x, ship.y), board_topleft)
        ship_pos = self.ships_descr[ship.size]['models'][ship.direction].get_rect()
        ship_pos.centerx = cx + self.ships_descr[ship.size]['offset'][ship.direction][0]
        ship_pos.centery = cy + self.ships_descr[ship.size]['offset'][ship.direction][1]
        s.blit(self.ships_descr[ship.size]['models'][ship.direction], ship_pos)

    def _draw_ship_descr(self, s):
        if self.tmp_ship is None:
            return

        help_top = 300
        help_left = 550

        if self.tmp_ship is not None:
            ship = self.ships_descr[self.tmp_ship.size]
            ship['turntable'].draw(s)
            name_rect = ship['rname'].get_rect()
            name_rect.left = help_left
            name_rect.top = ship['turntable'].rect.bottom - 50
            s.blit(ship['rname'], name_rect)

            top = name_rect.top + 80
            for rline in ship['rdescr']:
                s.blit(rline, (help_left, top))
                top += 20

    def _draw_board_objects(self, s, board_topleft, board):
        tmp_drawn = False
        for ship in sorted(board.ships, key=lambda ship: ship.y):
            if self.tmp_ship is not None and not tmp_drawn and ship.y >= self.tmp_ship.y:
                self._draw_ship(s, self.tmp_ship, board_topleft)
                tmp_drawn = True
            self._draw_ship(s, ship, board_topleft)
        if self.tmp_ship is not None and not tmp_drawn:
            self._draw_ship(s, self.tmp_ship, board_topleft)

    def _draw_stats(self, s, board_topleft, board):
        update_rects = []

        H = 29
        topleft = (board_topleft[0], 614)
        if len(board.ships)>0:
            s.blit(self.ships_descr[2]["models"]["P"], (topleft[0] + -45, topleft[1] + H - 10))
        if len(board.ships)>1:
            s.blit(self.ships_descr[3]["models"]["P"], (topleft[0] +  40, topleft[1] + H - 10))
        if len(board.ships)>2:
            s.blit(self.ships_descr[3]["models"]["P"], (topleft[0] + 130, topleft[1] + H - 10))
        if len(board.ships)>3:
            s.blit(self.ships_descr[4]["models"]["P"], (topleft[0] + 230, topleft[1] + H - 10))
        if len(board.ships)>4:
            s.blit(self.ships_descr[5]["models"]["P"], (topleft[0] + 355, topleft[1] + H - 10))
        update_rects.append(pygame.Rect(topleft, (topleft[0], s.get_width())))

        return update_rects

    def _prepare_grid(self, s):
        ocean_player = pygame.image.load("gfx/ocean_player.png").convert_alpha()

        ocean_width = 480
        top_center_x=int(510*480/870)
        left_center_y=int(267*480/870)
        ocean_ratio  = ocean_player.get_width()/ocean_player.get_height()
        ocean_zoom   = ocean_width / ocean_player.get_width()
        #ocean_player = pygame.transform.smoothscale(ocean_player, (ocean_width, int(ocean_width/ocean_ratio)))

        ocean_size   = ocean_player.get_size()

        # top left border of a grid cell
        self.cell_tl      = top_center_x/10 * ocean_zoom, left_center_y/10 * ocean_zoom
        # top right border of a grid cell
        self.cell_tr      = ocean_size[0]/10 - self.cell_tl[0], ocean_size[1]/10 - self.cell_tl[1]
        self.cell_w       = self.cell_tl[0] + self.cell_tr[0]
        self.cell_h       = self.cell_tl[1] + self.cell_tr[1]
        cell_angle = -math.asin((self.cell_tl[0]-self.cell_w/2)/self.cell_w) / math.pi * 180 + 45

        self.ocean_player_rect = ocean_player.get_rect()
        self.ocean_player_rect.topleft = (self.cell_tr[0], 36)
        s.blit(ocean_player, self.ocean_player_rect)

        corner = (self.ocean_player_rect.left, self.ocean_player_rect.top)

        # Drawing grid and legend
        grid_font = pygame.font.Font(None, 19)
        labels_cols = []
        labels_rows = []
        for i in range(0, 10):
            labels_rows.append(grid_font.render(str(i+1), 1, WHITE))
            labels_cols.append(grid_font.render(chr(ord('A') + i), 1, WHITE))
        for l in range(0,10+1):
            pygame.draw.line(s, WHITE, self.coords_to_pos(( l, 0), corner), self.coords_to_pos((l, 11), corner), 1)
            pygame.draw.line(s, WHITE, self.coords_to_pos((-1, l), corner), self.coords_to_pos((10, l), corner), 1)
        for l in range(0, 10):
            cx,cy = self.coords_to_pos((l, 10), corner)
            crect = labels_cols[l].get_rect()
            crect.centerx = cx + self.cell_tr[0]
            crect.centery = cy + self.cell_tr[1]/4
            s.blit(pygame.transform.rotate(labels_cols[l], cell_angle), crect)

            rx,ry = self.coords_to_pos((-1, l), corner)
            rrect = labels_rows[l].get_rect()
            rrect.centerx = rx + self.cell_tl[0]*3/4 -3
            rrect.centery = ry + self.cell_tr[1]*0/4
            s.blit(pygame.transform.rotate(labels_rows[l], cell_angle), rrect)

    def reset(self):
        self.show_help = False
        self.tmp_ship = None
        self.hlp_boat_num = None
        self.tmp_ship = Ship(size=2, x=4, y=4, direction=Direction.SOUTH)
        self.debug = False

    def update(self, rnd, params):
        if rnd.state.options.get('direction', None) is None:
            rnd.state.options['direction'] = Direction.EAST
        if rnd.state.options.get('x', None) is None:
            rnd.state.options['x'] = 4
        if rnd.state.options.get('y', None) is None:
            rnd.state.options['y'] = 4

        if not rnd.state.options.get('auto', False) and len(rnd.player.ships) < len(rnd.ships) and (
                self.tmp_ship is None
                or self.tmp_ship.x != rnd.state.options['x']
                or self.tmp_ship.y != rnd.state.options['y']
                or self.tmp_ship.direction != rnd.state.options['direction']
                or self.tmp_ship.size != rnd.ships[len(rnd.player.ships)][1]
                ):
            tmp_ship_size = rnd.ships[len(rnd.player.ships)][1]
            self.tmp_ship = Ship(
                    rnd.state.options['x'],
                    rnd.state.options['y'],
                    rnd.state.options['direction'],
                    tmp_ship_size,
                    self.ships_descr[tmp_ship_size]['name']
                    )
            rnd.state.options['needs_redraw'] = True

        if not self.show_help and self.tmp_ship is not None and params.get('animations', True):
            self.ships_descr[self.tmp_ship.size]['turntable'].update()
            rnd.state.options['needs_redraw'] = True

    def draw(self, s, board):
        update_rects = []

        s.blit(self.bg, (0, 0))
        update_rects.append(s.get_rect())

        self._draw_board_objects(s, self.ocean_player_rect.topleft, board)

        if not self.show_help:
            s.blit(self.menu_items[0][0], self.menu_rects[0])
            self._draw_ship_descr(s)
        else:
            s.blit(self.help_pane, (0, 0))
            s.blit(self.menu_items[0][1], self.menu_rects[0])
        update_rects.append(self.menu_rects[0])

        update_rects += self._draw_stats(s, self.ocean_player_rect.topleft, board)

        return update_rects

    def input(self, event, rnd):
        # help button
        if event.type == pygame.MOUSEBUTTONDOWN and self.menu_rects[0].collidepoint(event.pos):
                self.show_help = not self.show_help
                rnd.state.options['needs_redraw'] = True
        # help key
        elif event.type == pygame.KEYDOWN and event.unicode == self.menu_items[0][2]:
                self.show_help = not self.show_help
                rnd.state.options['needs_redraw'] = True
        # boat mouse command and rotation
        elif event.type == pygame.MOUSEBUTTONDOWN and self.ocean_player_rect.collidepoint(event.pos):
            if event.button == 3:
                rnd.state.options['needs_redraw'] = True
                if rnd.state.options['direction'] == Direction.EAST:
                    rnd.state.options['direction'] = Direction.SOUTH
                elif rnd.state.options['direction'] == Direction.SOUTH:
                    rnd.state.options['direction'] = Direction.WEST
                elif rnd.state.options['direction'] == Direction.WEST:
                    rnd.state.options['direction'] = Direction.NORTH
                elif rnd.state.options['direction'] == Direction.NORTH:
                    rnd.state.options['direction'] = Direction.EAST
            elif event.button == 1:
                rnd.state.options['ready'] = True
                self.tmp_ship = None
            elif event.button == 2:
                rnd.state.options['auto'] = True
                self.tmp_ship = None
        # boat mouse position
        elif event.type == pygame.MOUSEMOTION and self.ocean_player_rect.collidepoint(event.pos):
                x,y = self.pos_to_coords(event.pos, self.ocean_player_rect.topleft)
                try:
                    tmp_ship_size = rnd.ships[len(rnd.player.ships)][1]
                    if x>= 0 and y >= 0 and x < rnd.width and y < rnd.width and (
                            rnd.state.options['direction'] == Direction.NORTH and y >= tmp_ship_size-1
                            or rnd.state.options['direction'] == Direction.SOUTH and y <= rnd.width - tmp_ship_size
                            or rnd.state.options['direction'] == Direction.EAST  and x <= rnd.width - tmp_ship_size
                            or rnd.state.options['direction'] == Direction.WEST  and x >= tmp_ship_size-1
                            ):
                        rnd.state.options['x'] = x
                        rnd.state.options['y'] = y
                except:
                    pass
        # boat keyboard command and position
        elif event.type == pygame.KEYDOWN:
            # rotation
            if event.unicode == 'r':
                if   rnd.state.options['direction'] == Direction.EAST:
                    rnd.state.options['direction']   = Direction.SOUTH
                elif rnd.state.options['direction'] == Direction.SOUTH:
                    rnd.state.options['direction']   = Direction.WEST
                elif rnd.state.options['direction'] == Direction.WEST:
                    rnd.state.options['direction']   = Direction.NORTH
                elif rnd.state.options['direction'] == Direction.NORTH:
                    rnd.state.options['direction']   = Direction.EAST
            # position and command
            elif event.key in [pygame.K_UP, pygame.K_DOWN, pygame.K_LEFT, pygame.K_RIGHT, pygame.K_RETURN, pygame.K_SPACE] and not self.debug:
                if self.tmp_ship is not None and not rnd.state.options.get('ready', False):
                    rnd.state.options['needs_redraw'] = True
                    if event.key == pygame.K_UP:
                        if  self.tmp_ship.ymin > 0:
                            self.tmp_ship.y    -= 1
                            self.tmp_ship.ymin -= 1
                            self.tmp_ship.ymax -= 1
                            rnd.state.options['y'] = self.tmp_ship.y
                    elif event.key == pygame.K_DOWN:
                        if  self.tmp_ship.ymax < 9:
                            self.tmp_ship.y    += 1
                            self.tmp_ship.ymin += 1
                            self.tmp_ship.ymax += 1
                            rnd.state.options['y'] = self.tmp_ship.y
                    elif event.key == pygame.K_LEFT:
                        if  self.tmp_ship.xmin > 0:
                            self.tmp_ship.x    -= 1
                            self.tmp_ship.xmin -= 1
                            self.tmp_ship.xmax -= 1
                            rnd.state.options['x'] = self.tmp_ship.x
                    elif event.key == pygame.K_RIGHT:
                        if  self.tmp_ship.xmax < 9:
                            self.tmp_ship.x    += 1
                            self.tmp_ship.xmin += 1
                            self.tmp_ship.xmax += 1
                            rnd.state.options['x'] = self.tmp_ship.x
                    elif event.key == pygame.K_RETURN:
                        rnd.state.options['ready'] = True
                        self.tmp_ship = None
                    elif event.key == pygame.K_SPACE:
                        rnd.state.options['auto'] = True
                        self.tmp_ship = None
            # debug
            elif event.unicode == 'd':
                self.debug = not self.debug
            elif event.key in [pygame.K_UP, pygame.K_DOWN, pygame.K_LEFT, pygame.K_RIGHT, pygame.K_RETURN] and self.debug:
                if self.tmp_ship is not None:
                    rnd.state.options['needs_redraw'] = True
                    ox,oy = self.ships_descr[self.tmp_ship.size]['offset'][self.tmp_ship.direction]
                    if event.key == pygame.K_UP:
                        ox,oy = ox, oy-1
                    elif event.key == pygame.K_DOWN:
                        ox,oy = ox, oy+1
                    elif event.key == pygame.K_LEFT:
                        ox,oy = ox-1, oy
                    elif event.key == pygame.K_RIGHT:
                        ox,oy = ox+1, oy
                    elif event.key == pygame.K_RETURN:
                        print(self.ships_descr[self.tmp_ship.size]['offset'])
                    self.ships_descr[self.tmp_ship.size]['offset'][self.tmp_ship.direction] = ox,oy

class GuiBattleShips:
    def coords_to_pos(self, coords, topleft):
        y,x = coords
        a = self.cell_tr[0]
        b = self.cell_tl[0]
        c = self.cell_tr[1]
        d = -self.cell_tl[1]

        bx = x*a + y*b
        by = x*c + y*d

        by += 10*self.cell_tl[1]

        # ocean offset
        bx += topleft[0]
        by += topleft[1]
        return (int(bx), int(by))

    def pos_to_coords(self, pos, topleft):
        bx, by = pos
        # ocean offset
        bx -= topleft[0]
        by -= topleft[1]

        by -= 10*self.cell_tl[1]

        a = -self.cell_tl[1]
        b = -self.cell_tl[0]
        c = -self.cell_tr[1]
        d =  self.cell_tr[0]

        det = a*d-b*c
        x, y = (a*bx + b*by)/det, (c*bx + d*by)/det
        x, y = math.floor(x), math.floor(y)

        return (y,x)

    def __init__(self):
        self.reset()

        bg = pygame.Surface((1024, 768))

        bigboard_rect     = pygame.Rect((0, 0), (bg.get_width(), 614))
        hitboard_rect     = pygame.Rect((0, 614), (bg.get_width(), bg.get_height()-614))
        pygame.draw.rect(bg, DARKGRAY,  bigboard_rect)
        pygame.draw.rect(bg, LIGHTGRAY, hitboard_rect)

        font_player = pygame.font.Font(FONT, 16)
        label_player = font_player.render(_("Player"), 1, WHITE)
        label_enemy  = font_player.render(_("Enemy"), 1, WHITE)
        r = label_enemy.get_rect()
        r.topright = (bg.get_width()-10, 10)
        bg.blit(label_player, (10, 10))
        bg.blit(label_enemy, r)

        self._prepare_grids(bg)

        pygame.draw.line(bg, MEDIUMGRAY, (10, hitboard_rect.bottom - 35), (10 + hitboard_rect.width/2 - 10, hitboard_rect.bottom - 35), 1)

        self.bg = bg

        self._prepare_ship_descr()

        self.miss = pygame.image.load("gfx/cross_white.png").convert_alpha()
        #self.miss = pygame.transform.smoothscale(self.miss, (int(self.cell_w/2), int(self.cell_h/2)))
        self.hit  = pygame.image.load("gfx/cross_orange.png").convert_alpha()
        #self.hit  = pygame.transform.smoothscale(self.hit, (int(self.cell_w/2), int(self.cell_h/2)))
        self.sunk = pygame.image.load("gfx/cross_red.png").convert_alpha()
        #self.sunk = pygame.transform.smoothscale(self.sunk, (int(self.cell_w/2), int(self.cell_h/2)))

        self.font_stats = pygame.font.Font(FONT, 20)

        self.labels_stats = {}
        self.labels_stats["shots"] = self.font_stats.render(_("Shots"), 1, WHITE)
        self.labels_stats["miss"]  = self.font_stats.render(_("Miss"), 1, WHITE)
        self.labels_stats["hit"]   = self.font_stats.render(_("Hit"), 1, WHITE)
        self.labels_stats["sunk"]  = self.font_stats.render(_("Sunk"), 1, BLACK)


        self.notif_rect   = pygame.Rect((0, 450), (bg.get_width(), 50))

    def _prepare_ship_descr(self):
        self.ships_descr = [None, None]
        for name, size, descr in [
                (_("The Anger"),         2, _("Every shipman is only grousing day and night, which drives the captain completely mad. There is one thing you can do to help him unwind, that is be an easy pray. I would advise you not to…")),
                (_("The Victory Saber"), 3, _("They were twin brothers, incredibly talentuous swashbucklers, and the fell for a mysterious lady. She said only the most fearsome pirate would be her match. Will you try their cruelty?")),
                (_("The Pirates Storm"), 4, _("This vessel is known for its sneaky attacks during storms. Manœuvrable and fast, no ship can escape her in the raging waters. If the weather is taking a turn for the worse, be sure that you're alone…")),
                (_("The Princess"),      5, _("Rather than the vessel, 'The Princess' refers to its heartless captain. Famous buccaneer, she is said to be as rich as a king. Well, if she meets you, maybe she well become even richer…")),
                ]:
            ship = {}
            ship['name']        = name
            ship['descr']       = descr
            models = {}
            for direction,letter in [(Direction.NORTH, "N"), (Direction.SOUTH, "S"), (Direction.EAST, "E"), (Direction.WEST, "W"), ("P", "P")]:
                view = pygame.image.load("gfx/boat_"+str(size)+"_"+letter+".png").convert_alpha()
                #view = pygame.transform.smoothscale(view, (int(5*self.cell_tl[0]),int(5*self.cell_tl[0])))
                models[direction] = view
            ship['models'] = models
            self.ships_descr.append(ship)

        self.ships_descr[2]['offset'] = {
                Direction.NORTH: ( 17,-17),
                Direction.SOUTH: ( 30, -7),
                Direction.EAST:  ( 36,-23),
                Direction.WEST:  ( 10,-14),
                'P': ( 0,  0),
                }
        self.ships_descr[3]['offset'] = {
                Direction.NORTH: ( 10,-25),
                Direction.SOUTH: ( 41,  4),
                Direction.EAST:  ( 47,-27),
                Direction.WEST:  ( -5, -5),
                'P': ( 0,  0),
                }
        self.ships_descr[4]['offset'] = {
                Direction.NORTH: ( -2,-36),
                Direction.SOUTH: ( 53, 16),
                Direction.EAST:  ( 69,-36),
                Direction.WEST:  (-16,  1),
                'P': ( 0,  0),
                }
        self.ships_descr[5]['offset'] = {
                Direction.NORTH: ( -9, -45),
                Direction.SOUTH: ( 66,  32),
                Direction.EAST:  ( 81, -44),
                Direction.WEST:  (-36,  11),
                'P': ( 0,  0),
                }

    def _prepare_grids(self, s):
        ocean_player = pygame.image.load("gfx/ocean_player.png").convert_alpha()
        ocean_cpu    = pygame.image.load("gfx/ocean_cpu.png").convert_alpha()

        ocean_width = 480
        top_center_x=int(510*480/870)
        left_center_y=int(267*480/870)
        ocean_ratio  = ocean_player.get_width()/ocean_player.get_height()
        ocean_zoom   = ocean_width / ocean_player.get_width()
        #ocean_player = pygame.transform.smoothscale(ocean_player, (ocean_width, int(ocean_width/ocean_ratio)))
        #ocean_cpu    = pygame.transform.smoothscale(ocean_cpu,    (ocean_width, int(ocean_width/ocean_ratio)))

        ocean_size   = ocean_player.get_size()

        # top left border of a grid cell
        self.cell_tl      = top_center_x/10 * ocean_zoom, left_center_y/10 * ocean_zoom
        # top right border of a grid cell
        self.cell_tr      = ocean_size[0]/10 - self.cell_tl[0], ocean_size[1]/10 - self.cell_tl[1]
        self.cell_w       = self.cell_tl[0] + self.cell_tr[0]
        self.cell_h       = self.cell_tl[1] + self.cell_tr[1]
        cell_angle = -math.asin((self.cell_tl[0]-self.cell_w/2)/self.cell_w) / math.pi * 180 + 45

        self.ocean_player_rect = ocean_player.get_rect()
        self.ocean_cpu_rect    = ocean_cpu.get_rect()
        self.ocean_player_rect.topleft = (self.cell_tr[0], 36)
        self.ocean_cpu_rect.topleft    = (512, 36)
        s.blit(ocean_player, self.ocean_player_rect)
        s.blit(ocean_cpu, self.ocean_cpu_rect)

        for corner in [self.ocean_player_rect.topleft, self.ocean_cpu_rect.topleft]:
            # Drawing grid and legend
            grid_font = pygame.font.Font(None, 19)
            labels_cols = []
            labels_rows = []
            for i in range(0, 10):
                labels_rows.append(grid_font.render(str(i+1), 1, WHITE))
                labels_cols.append(grid_font.render(chr(ord('A') + i), 1, WHITE))
            for l in range(0,10+1):
                pygame.draw.line(s, WHITE, self.coords_to_pos(( l, 0), corner), self.coords_to_pos((l, 11), corner), 1)
                pygame.draw.line(s, WHITE, self.coords_to_pos((-1, l), corner), self.coords_to_pos((10, l), corner), 1)
            for l in range(0, 10):
                cx,cy = self.coords_to_pos((l, 10), corner)
                crect = labels_cols[l].get_rect()
                crect.centerx = cx + self.cell_tr[0]
                crect.centery = cy + self.cell_tr[1]/4
                s.blit(pygame.transform.rotate(labels_cols[l], cell_angle), crect)

                rx,ry = self.coords_to_pos((-1, l), corner)
                rrect = labels_rows[l].get_rect()
                rrect.centerx = rx + self.cell_tl[0]*3/4 -3
                rrect.centery = ry + self.cell_tr[1]*0/4
                s.blit(pygame.transform.rotate(labels_rows[l], cell_angle), rrect)

    def _draw_stats(self, s, topleft, board, other, show_hits=True):
        update_rects = []

        h = 24
        H = h+5

        topleft = topleft[0], 614
        label_stats_width = max([self.labels_stats[i].get_width() for i in ["shots", "miss", "hit", "sunk"]])

        s.blit(self.labels_stats["shots"], (topleft[0], topleft[1] - 3*H))
        s.blit(self.labels_stats["miss"],  (topleft[0], topleft[1] - 2*H))
        s.blit(self.labels_stats["hit"],   (topleft[0], topleft[1] - 1*H))
        s.blit(self.labels_stats["sunk"],  (topleft[0], topleft[1] - 0*H + 4))

        rect_shots  = pygame.Rect((topleft[0] + label_stats_width + 15,            topleft[1] - 3*H),     (h*2, h))
        rect_miss   = pygame.Rect((topleft[0] + label_stats_width + 15,            topleft[1] - 2*H),     (h*2, h))
        rect_hit    = pygame.Rect((topleft[0] + label_stats_width + 15,            topleft[1] - 1*H),     (h*2, h))
        rect_sunk   = pygame.Rect((topleft[0] + label_stats_width + 15,            topleft[1] - 0*H + 4), (h*2, h))
        rect_miss_c = pygame.Rect((topleft[0] + label_stats_width + 15 + 2*h + 10, topleft[1] - 2*H + 2), (h*2, h))
        rect_hit_c  = pygame.Rect((topleft[0] + label_stats_width + 15 + 2*h + 10, topleft[1] - 1*H + 2), (h*2, h))
        rect_sunk_c = pygame.Rect((topleft[0] + label_stats_width + 15 + 2*h + 10, topleft[1] - 0*H + 6), (h*2, h))
        pygame.draw.rect(s, WHITE, rect_shots)
        pygame.draw.rect(s, WHITE, rect_miss)
        pygame.draw.rect(s, WHITE, rect_hit)
        pygame.draw.rect(s, BLACK, rect_sunk)
        s.blit(self.miss, rect_miss_c)
        s.blit(self.hit,  rect_hit_c)
        s.blit(self.sunk, rect_sunk_c)
        label_shots = self.font_stats.render(str(len(other.miss)+len(other.hits)), 1, BLACK)
        label_miss  = self.font_stats.render(str(len(other.miss)), 1, BLACK)
        label_hit   = self.font_stats.render(str(len(other.hits)), 1, BLACK)
        label_sunk  = self.font_stats.render(str(len(other.sunk)), 1, WHITE)
        pos_shots = label_shots.get_rect()
        pos_miss  = label_miss.get_rect()
        pos_hit   = label_hit.get_rect()
        pos_sunk   = label_sunk.get_rect()
        pos_shots.center = rect_shots.center
        pos_miss.center = rect_miss.center
        pos_hit.center = rect_hit.center
        pos_sunk.center = rect_sunk.center
        s.blit(label_shots, pos_shots)
        s.blit(label_miss, pos_miss)
        s.blit(label_hit, pos_hit)
        s.blit(label_sunk, pos_sunk)
        update_rects+=[pos_shots, pos_miss, pos_sunk]

        s.blit(self.ships_descr[2]["models"]["P"], (topleft[0] + -45, topleft[1] + H - 10))
        s.blit(self.ships_descr[3]["models"]["P"], (topleft[0] +  40, topleft[1] + H - 10))
        s.blit(self.ships_descr[3]["models"]["P"], (topleft[0] + 130, topleft[1] + H - 10))
        s.blit(self.ships_descr[4]["models"]["P"], (topleft[0] + 230, topleft[1] + H - 10))
        s.blit(self.ships_descr[5]["models"]["P"], (topleft[0] + 355, topleft[1] + H - 10))

        topleft = (topleft[0], topleft[1] + 155)

        if show_hits or board.ships[0].get_health()==0:
            if board.ships[0].get_health() == 0:
                icon = self.sunk
            else:
                icon = self.hit
            if board.ships[0].health[0] == 0:
                s.blit(icon, (topleft[0] +   0 + (icon.get_width() + 1)*0, topleft[1] - 28))
            if board.ships[0].health[1] == 0:
                s.blit(icon, (topleft[0] +   0 + (icon.get_width() + 1)*1, topleft[1] - 28))
        if show_hits or board.ships[1].get_health()==0:
            if board.ships[1].get_health() == 0:
                icon = self.sunk
            else:
                icon = self.hit
            if board.ships[1].health[0] == 0: 
                s.blit(icon, (topleft[0] +  70 + (icon.get_width() + 1)*0, topleft[1] - 28))
            if board.ships[1].health[1] == 0: 
                s.blit(icon, (topleft[0] +  70 + (icon.get_width() + 1)*1, topleft[1] - 28))
            if board.ships[1].health[2] == 0: 
                s.blit(icon, (topleft[0] +  70 + (icon.get_width() + 1)*2, topleft[1] - 28))

        if show_hits or board.ships[2].get_health()==0:
            if board.ships[2].get_health() == 0:
                icon = self.sunk
            else:
                icon = self.hit
            if board.ships[2].health[0] == 0: 
                s.blit(icon, (topleft[0] + 160 + (icon.get_width() + 1)*0, topleft[1] - 28))
            if board.ships[2].health[1] == 0: 
                s.blit(icon, (topleft[0] + 160 + (icon.get_width() + 1)*1, topleft[1] - 28))
            if board.ships[2].health[2] == 0: 
                s.blit(icon, (topleft[0] + 160 + (icon.get_width() + 1)*2, topleft[1] - 28))

        if show_hits or board.ships[3].get_health()==0:
            if board.ships[3].get_health() == 0:
                icon = self.sunk
            else:
                icon = self.hit
            if board.ships[3].health[0] == 0: 
                s.blit(icon, (topleft[0] + 250 + (icon.get_width() + 1)*0, topleft[1] - 28))
            if board.ships[3].health[1] == 0: 
                s.blit(icon, (topleft[0] + 250 + (icon.get_width() + 1)*1, topleft[1] - 28))
            if board.ships[3].health[2] == 0: 
                s.blit(icon, (topleft[0] + 250 + (icon.get_width() + 1)*2, topleft[1] - 28))
            if board.ships[3].health[3] == 0: 
                s.blit(icon, (topleft[0] + 250 + (icon.get_width() + 1)*3, topleft[1] - 28))

        if show_hits or board.ships[4].get_health()==0:
            if board.ships[4].get_health() == 0:
                icon = self.sunk
            else:
                icon = self.hit
            if board.ships[4].health[0] == 0: 
                s.blit(icon, (topleft[0] + 365 + (icon.get_width() + 1)*0, topleft[1] - 28))
            if board.ships[4].health[1] == 0: 
                s.blit(icon, (topleft[0] + 365 + (icon.get_width() + 1)*1, topleft[1] - 28))
            if board.ships[4].health[2] == 0: 
                s.blit(icon, (topleft[0] + 365 + (icon.get_width() + 1)*2, topleft[1] - 28))
            if board.ships[4].health[3] == 0: 
                s.blit(icon, (topleft[0] + 365 + (icon.get_width() + 1)*3, topleft[1] - 28))
            if board.ships[4].health[4] == 0: 
                s.blit(icon, (topleft[0] + 365 + (icon.get_width() + 1)*4, topleft[1] - 28))

        update_rects.append(pygame.Rect(topleft, (s.get_width(), s.get_height())))

        return update_rects



    def _draw_ship(self, s, ship, board_topleft):
        cx, cy = self.coords_to_pos((ship.x, ship.y), board_topleft)
        ship_pos = self.ships_descr[ship.size]['models'][ship.direction].get_rect()
        ship_pos.centerx = cx + self.ships_descr[ship.size]['offset'][ship.direction][0]
        ship_pos.centery = cy + self.ships_descr[ship.size]['offset'][ship.direction][1]
        s.blit(self.ships_descr[ship.size]['models'][ship.direction], ship_pos)

    def _draw_board_objects(self, s, board_topleft, board, draw_ships=True):
        update_rects = []
        # Drawing ploufs
        for (mx, my) in board.miss:
            a = [e[0] for e in self.event_anims if e[1]==(mx, my) and e[2]==board_topleft]
            if len(a) == 0:
                x, y = self.coords_to_pos((mx, my),board_topleft)
                cross_pos = self.miss.get_rect()
                cross_pos.center = (x + (self.cell_tl[0] + self.cell_tr[0])/2, y + (self.cell_tr[1]-self.cell_tl[1])/2)
                s.blit(self.miss, cross_pos)
                update_rects+=cross_pos
            else:
                if a[0].index is not None:
                    rl = a[0].draw(s)
                    update_rects += rl
        # Drawing ships
        if draw_ships:
            for ship in sorted(board.ships, key=lambda ship: ship.y):
                self._draw_ship(s, ship, board_topleft)

        # Sunk positions
        sunk_coords = []
        for ship in board.ships:
            if ship.get_health() == 0:
                for i in range(ship.xmin, ship.xmax+1):
                    for j in range(ship.ymin, ship.ymax+1):
                        sunk_coords.append((i,j))

        # Drawing boums
        for (hx, hy) in board.hits:
            a = [e[0] for e in self.event_anims if e[1]==(hx, hy) and e[2]==board_topleft]
            if len(a) == 0:
                x, y = self.coords_to_pos((hx, hy), board_topleft)
                cross_pos = self.hit.get_rect()
                cross_pos.center = (x + (self.cell_tl[0] + self.cell_tr[0])/2, y + (self.cell_tr[1]-self.cell_tl[1])/2)
                if (hx, hy) in sunk_coords:
                    s.blit(self.sunk, cross_pos)
                else:
                    s.blit(self.hit, cross_pos)
                update_rects+=cross_pos
            else:
                if a[0].index is not None:
                    rl = a[0].draw(s)
                    update_rects += rl

        return update_rects

    def reset(self):
        self.event_anims  = []
        self.prev_player_hits = []
        self.prev_player_miss = []
        self.prev_cpu_hits    = []
        self.prev_cpu_miss    = []
        self.debug = False
        self.draw_enemy_ships = False
        self.aiming = 4,4

    def update(self, rnd, params):
        if rnd.state.options.get('x', None) is None:
            rnd.state.options['x'] = self.aiming[0]
        if rnd.state.options.get('y', None) is None:
            rnd.state.options['y'] = self.aiming[1]
        self.aiming = (rnd.state.options['x'], rnd.state.options['y'])

        self.draw_enemy_ships = params['ai'] == AI.CHEAT
        for (old, new, topleft) in [
                (self.prev_player_hits, rnd.player.hits, self.ocean_player_rect.topleft),
                (self.prev_cpu_hits,    rnd.cpu.hits,    self.ocean_cpu_rect.topleft),
                ]:
            for (hx, hy) in new[len(old):]:
                (bx, by) = self.coords_to_pos((hx, hy), topleft)
                if params.get('animations', True):
                    self.event_anims.append((
                        #FxHit(pos=(bx, by), loop=False, resize_width=int(self.cell_w*2), offset=(self.cell_tl[0], -self.cell_tl[1] - 40)),
                        FxHit(pos=(bx, by), loop=False, offset=(self.cell_tl[0], -self.cell_tl[1] - 40)),
                        (hx, hy),
                        topleft
                        ))
                if params.get('sfx', True):
                    sound = pygame.mixer.Sound("sfx/hit.ogg")
                    sound.play()

        self.prev_player_hits = rnd.player.hits.copy()
        self.prev_cpu_hits    = rnd.cpu.hits.copy()

        for (old, new, topleft) in [
                (self.prev_player_miss, rnd.player.miss, self.ocean_player_rect.topleft),
                (self.prev_cpu_miss,    rnd.cpu.miss,    self.ocean_cpu_rect.topleft),
                ]:
            for (hx, hy) in new[len(old):]:
                (bx, by) = self.coords_to_pos((hx, hy), topleft)
                if params.get('animations', True):
                    self.event_anims.append((
                        #FxMiss(pos=(bx, by), loop=False, resize_width=int(self.cell_w*1), offset=(self.cell_tl[0], 0)),
                        FxMiss(pos=(bx, by), loop=False, offset=(self.cell_tl[0]-3, +3)),
                        (hx, hy),
                        topleft
                        ))
                if params.get('sfx', True):
                    sound = pygame.mixer.Sound("sfx/miss.ogg")
                    sound.play()

        self.prev_player_miss = rnd.player.miss.copy()
        self.prev_cpu_miss    = rnd.cpu.miss.copy()

        for (a, pos, topleft) in self.event_anims:
            a.update()

        if len(self.event_anims) > 0:
            rnd.state.options['needs_redraw'] = True

        self.event_anims = [ (a, (x,y), topleft) for (a, (x,y), topleft) in self.event_anims if not a.finished]


    def draw(self, s, player, cpu):
        update_rects = []

        s.blit(self.bg, (0, 0))
        update_rects.append(s.get_rect())

        x,y = self.aiming
        pygame.draw.polygon(s, ORANGE, [ 
            self.coords_to_pos((x, y),     self.ocean_cpu_rect.topleft),
            self.coords_to_pos((x+1, y),   self.ocean_cpu_rect.topleft),
            self.coords_to_pos((x+1, y+1), self.ocean_cpu_rect.topleft),
            self.coords_to_pos((x, y+1),   self.ocean_cpu_rect.topleft),
            ], 3)

        self._draw_board_objects(s, self.ocean_player_rect.topleft, player)
        self._draw_board_objects(s, self.ocean_cpu_rect.topleft, cpu, draw_ships=self.draw_enemy_ships)

        self._draw_stats(s, self.ocean_player_rect.topleft, player, cpu)
        self._draw_stats(s, (self.ocean_cpu_rect.topleft[0] + 20, self.ocean_cpu_rect.topleft[1]), cpu, player, show_hits=self.draw_enemy_ships)

        return update_rects


    def input(self, event, rnd):
            # mouse command
            if event.type == pygame.MOUSEBUTTONDOWN and self.ocean_cpu_rect.collidepoint(event.pos):
                if event.button == 1:
                    x,y = self.pos_to_coords(event.pos, self.ocean_cpu_rect.topleft)
                    if 0 <= x and 0 <= y and x < rnd.width and y < rnd.width:
                        rnd.state.options['needs_redraw'] = True
                        rnd.state.options['x'] = x
                        rnd.state.options['y'] = y
                        rnd.state.options['ready'] = True
                elif event.button == 2:
                    rnd.state.options['auto'] = True
                    rnd.state.options['needs_redraw'] = True
            # mouse position
            elif event.type == pygame.MOUSEMOTION and self.ocean_cpu_rect.collidepoint(event.pos):
                    x,y = self.pos_to_coords(event.pos, self.ocean_cpu_rect.topleft)
                    if x>= 0 and y >= 0 and x < rnd.width and y < rnd.width:
                        rnd.state.options['x'] = x
                        rnd.state.options['y'] = y
                        rnd.state.options['needs_redraw'] = True

            # keyboard command and position
            elif event.type == pygame.KEYDOWN and event.key in [pygame.K_UP, pygame.K_DOWN, pygame.K_LEFT, pygame.K_RIGHT, pygame.K_RETURN, pygame.K_SPACE] and not self.debug:
                if event.key == pygame.K_RETURN:
                    rnd.state.options['ready'] = True
                    rnd.state.options['needs_redraw'] = True
                elif event.key == pygame.K_SPACE:
                    rnd.state.options['auto'] = True
                    rnd.state.options['needs_redraw'] = True
                elif event.key in [pygame.K_UP, pygame.K_DOWN, pygame.K_LEFT, pygame.K_RIGHT]:
                    rnd.state.options['needs_redraw'] = True
                    x, y = rnd.state.options['x'], rnd.state.options['y']
                    if event.key == pygame.K_UP and rnd.state.options['y'] > 0:
                            rnd.state.options['y'] -= 1
                    elif event.key == pygame.K_DOWN and rnd.state.options['y'] < rnd.width - 1:
                            rnd.state.options['y'] += 1
                    elif event.key == pygame.K_LEFT and rnd.state.options['x'] > 0:
                            rnd.state.options['x'] -= 1
                    elif event.key == pygame.K_RIGHT and rnd.state.options['x'] < rnd.width - 1:
                            rnd.state.options['x'] += 1

class GuiGameOver:
    def __init__(self):
        gameover  = pygame.image.load("gfx/gameover.jpg").convert()

        self.bg = gameover

        menu_font = pygame.font.Font(FONT, 20)
        self.menu_items = []
        self.menu_items.append((
            menu_font.render(_("Continue"), 1, ORANGE),
            menu_font.render(_("Continue"), 1, WHITE),
            'm',
            Stage.CREDITS,
            ))
        self.menu_bottomright = (994, 738)

        self.menu_rects = []
        x,y = self.menu_bottomright
        for (n,(active, inactive, key, stage)) in enumerate(self.menu_items):
            rect = active.get_rect()
            rect.bottomright = (x,y)
            self.menu_rects.append(rect)
            y += rect.height + 10
            x += 20

        font_gameover = pygame.font.Font(FONT, 20)
        self.labels = {}
        self.labels["win"]  = font_gameover.render(_("You have won."), 1, WHITE)
        self.labels["lose"] = font_gameover.render(_("You have lost."), 1, WHITE)

        self.font_score = pygame.font.Font(FONT, 12)
        self.pos_score = (512, 658)

        label_player = self.font_score.render(_("Player"), 1, WHITE)
        (x, y) = self.pos_score
        x -= 120
        y -=  20
        p_rect = label_player.get_rect()
        p_rect.center = (x,y)

        label_enemy  = self.font_score.render(_("Enemy"), 1, WHITE)
        (x, y) = self.pos_score
        x += 120
        y -=  20
        c_rect = label_enemy.get_rect()
        c_rect.center = (x,y)

        self.bg.blit(label_player, p_rect)
        self.bg.blit(label_enemy,  c_rect)

        self.coins = pygame.image.load("gfx/winner_coins_light.png")

        self.reset()

    def reset(self):
        self.menu_index = 0

    def draw(self, s, rnd):
        if rnd.state.options.get('victory', None) is None:
            rnd.state.options['victory'] = False

        update_rects = []

        s.blit(self.bg, (0, 0))
        update_rects.append(s.get_rect())

        for (n, (active, inactive, key, stage)) in enumerate(self.menu_items):
            if self.menu_index == n:
                item = active
            else:
                item = inactive
            s.blit(item, self.menu_rects[n])
            update_rects.append(self.menu_rects[n])

        outcome_center = (s.get_width()//2, s.get_height()//2 + 160)

        if rnd.state.options.get('victory', False):
            self.labels["win"]
            outcome_rect = self.labels["win"].get_rect()
            outcome_rect.center = outcome_center
            s.blit(self.labels["win"], outcome_rect)
            update_rects.append(outcome_rect)
            coins_r = self.coins.get_rect()
            coins_r.midbottom = (self.pos_score[0] - 120, self.pos_score[1] - 15)
            s.blit(self.coins, coins_r)
            update_rects.append(coins_r)

        elif not rnd.state.options.get('victory', True):
            self.labels["lose"]
            outcome_rect = self.labels["lose"].get_rect()
            outcome_rect.center = outcome_center
            s.blit(self.labels["lose"], outcome_rect)
            update_rects.append(outcome_rect)
            coins_r = self.coins.get_rect()
            coins_r.midbottom = (self.pos_score[0] + 120, self.pos_score[1] - 15)
            s.blit(self.coins, coins_r)
            update_rects.append(coins_r)

        for (board, midtop) in [(rnd.cpu, (self.pos_score[0] - 120, self.pos_score[1] + 5)), (rnd.player, (self.pos_score[0] + 120, self.pos_score[1] + 5))]:
            (x, y) = midtop
            for score in [
                    str(len(board.hits) + len(board.miss)) + " " + _("Shots"),
                    str(len(board.hits)) + " " + _("Hit"),
                    str(len(board.miss)) + " " + _("Miss"),
                    str(len(board.sunk)) + " " + _("Sunk"),
                    ]:
                label  = self.font_score.render(score, 1, WHITE)
                rect = label.get_rect()
                rect.midtop = (x, y+5)
                (x, y) = rect.midbottom
                s.blit(label, rect)
                update_rects.append(rect)

        return update_rects

    def update(self, rnd, params):
        pass

    def input(self, event, rnd):
        if event.type == pygame.MOUSEBUTTONDOWN:
            if self.menu_rects[0].collidepoint(event.pos):
                rnd.state.options['skip'] = True
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_RETURN:
                rnd.state.options['skip'] = True

class PyGameUI:
    def __init__(self):
        (language, encoding) = locale.getdefaultlocale()

        try:
            locale_dir = gettext.bindtextdomain("capbattleship")
        except:
            locale_dir = None
        finally:
            if locale_dir is None:
                locale_dir = "locale"

        try:
            tr = gettext.translation("capbattleship", localedir=locale_dir, languages=[language[0:2].lower()])
        except:
            tr = gettext.translation("capbattleship", localedir=locale_dir, languages=["en"])
        tr.install()

        # workaround: disable sound if sdl1 is used
        (a,b,c) = pygame.version.vernum
        if a == 1:
            pygame.mixer.pre_init(frequency=44100, size=-16, channels=2, buffer=512)
        pygame.init()
        icon = pygame.image.load("gfx/icon_256px.png")
        pygame.display.set_icon(icon)

        self.clock = pygame.time.Clock()
        self.screen = pygame.display.set_mode((1024,768))

        pygame.display.set_caption(_("CAP - Pirate Battleship"))

        self.gui_menu = GuiMenu()
        self.gui_credits = GuiCredits()
        self.gui_options = GuiOptions()
        self.gui_setupships = GuiSetupShips()
        self.gui_battleships = GuiBattleShips()
        self.gui_gameover = GuiGameOver()

        self.font_notifs = pygame.font.Font(FONT, 18)
        self.labels_notifs = {}
        self.labels_notifs["CPU hit you."]    = self.font_notifs.render(_("CPU hit you."), 1, WHITE)
        self.labels_notifs["CPU missed you."] = self.font_notifs.render(_("CPU missed you."), 1, WHITE)
        self.labels_notifs["CPU sunk you."]   = self.font_notifs.render(_("CPU sunk you."), 1, WHITE)

        self.labels_notifs["You hit CPU."]    = self.font_notifs.render(_("You hit CPU."), 1, WHITE)
        self.labels_notifs["You missed CPU."] = self.font_notifs.render(_("You missed CPU."), 1, WHITE)
        self.labels_notifs["You sunk CPU."]   = self.font_notifs.render(_("You sunk CPU."), 1, WHITE)

        self.labels_notifs["Already shot there."]   = self.font_notifs.render(_("Already shot there."), 1, WHITE)

        self.labels_notifs["CPU starts."]     = self.font_notifs.render(_("CPU starts."), 1, WHITE)
        self.labels_notifs["You start."]      = self.font_notifs.render(_("You start."), 1, WHITE)

        self.bgsound = pygame.mixer.Sound("music/scottbuckley_titan.ogg")
        self.bgsound.set_volume(0.25)
        self.bgsound_start = None

        self.fullscreen = False

    def draw_notification(self, s, message):
        notif_msg = self.labels_notifs[message]
        notif_rect = notif_msg.get_rect()
        notif_rect.center = (512, 460)
        s.blit(notif_msg, notif_rect)
        redraw_rect = notif_rect
        redraw_rect.left = 0
        redraw_rect.right = s.get_width()
        return [redraw_rect]

    def update(self, rnd):
        bgsound_param = rnd.params.get('music', None)
        if self.bgsound_start is None and (bgsound_param in [BgRepeat.ONCE, BgRepeat.LOOP] or bgsound_param == BgRepeat.MENU and rnd.state.stage in [Stage.MENU, Stage.CREDITS, Stage.OPTIONS]) or bgsound_param == BgRepeat.LOOP and self.bgsound_start is not None and pygame.time.get_ticks() - self.bgsound_start > self.bgsound.get_length()*1000:
            self.bgsound_start = pygame.time.get_ticks()
            self.bgsound.play()
        elif self.bgsound_start is not None and bgsound_param == BgRepeat.NONE or bgsound_param == BgRepeat.MENU and rnd.state.stage not in [Stage.MENU, Stage.CREDITS, Stage.OPTIONS]:
            self.bgsound_start = None
            self.bgsound.stop()
        if rnd.state.stage == Stage.MENU:
            self.gui_menu.update(rnd, rnd.params)
        if rnd.state.stage in [ Stage.PLAYER_SHIPS, Stage.CPU_SHIPS ]:
            self.gui_setupships.update(rnd, rnd.params)
        elif rnd.state.stage in [ Stage.PLAYER_TURN, Stage.CPU_TURN, Stage.NOTIFICATION ]:
            self.gui_battleships.update(rnd, rnd.params)
        elif rnd.state.stage == Stage.GAMEOVER:
            self.gui_gameover.update(rnd, rnd.params)
        elif rnd.state.stage == Stage.OPTIONS:
            self.gui_options.update(rnd, rnd.params)
        elif rnd.state.stage == Stage.CREDITS:
            self.gui_credits.update(rnd, rnd.params)

        if rnd.state.stage == rnd.State.Stage.NOTIFICATION:
            now = pygame.time.get_ticks()
            timeout = rnd.params.get('notif_timeout', None)
            t = rnd.state.options.get('timer', None)
            if t is None:
                rnd.state.options['timer'] = now
            elif timeout is not None and now - t > timeout:
                rnd.state.options['skip'] = True

    def draw(self, rnd):
        self.update(rnd)

        if rnd.params.get('fullscreen', False) and not self.fullscreen:
            pygame.display.quit()
            pygame.display.init()
            flags = pygame.FULLSCREEN
            self.screen = pygame.display.set_mode((1024, 768), flags)
            self.fullscreen = True
        elif not rnd.params.get('fullscreen', False) and self.fullscreen:
            pygame.display.quit()
            pygame.display.init()
            self.screen = pygame.display.set_mode((1024, 768))
            self.fullscreen = False

        update_rects = []

        if rnd.state.stage == rnd.State.Stage.MENU:
            if rnd.state.options.get('needs_redraw', True):
                update_rects += self.gui_menu.draw(self.screen)
                rnd.state.options['needs_redraw'] = False

        elif rnd.state.stage == rnd.State.Stage.CREDITS:
            if rnd.state.options.get('needs_redraw', True):
                update_rects += self.gui_credits.draw(self.screen)
                rnd.state.options['needs_redraw'] = False

        elif rnd.state.stage == rnd.State.Stage.OPTIONS:
            if rnd.state.options.get('needs_redraw', True):
                update_rects += self.gui_options.draw(self.screen)
                rnd.state.options['needs_redraw'] = False

        elif rnd.state.stage in [rnd.State.Stage.PLAYER_SHIPS, rnd.State.Stage.CPU_SHIPS]:
            if rnd.state.options.get('needs_redraw', True):
                update_rects += self.gui_setupships.draw(self.screen, rnd.player)
                rnd.state.options['needs_redraw'] = False

        elif rnd.state.stage in [rnd.State.Stage.PLAYER_TURN, rnd.State.Stage.CPU_TURN, rnd.State.Stage.NOTIFICATION]:
            if rnd.state.options.get('needs_redraw', True):
                update_rects += self.gui_battleships.draw(self.screen, rnd.player, rnd.cpu)
                if rnd.state.options.get("message", None) is not None:
                    update_rects+=self.draw_notification(self.screen, rnd.state.options["message"])
                rnd.state.options['needs_redraw'] = False

        elif rnd.state.stage == rnd.State.Stage.GAMEOVER:
            if rnd.state.options.get('needs_redraw', True):
                update_rects+=self.gui_gameover.draw(self.screen, rnd)
                rnd.state.options['needs_redraw'] = False
        pygame.display.update(update_rects)
        self.clock.tick(10)

    def read_events(self, rnd):
        for event in pygame.event.get():

            if event.type == pygame.QUIT or (event.type==pygame.KEYDOWN and event.unicode=='q'):
                rnd.state.switch(Stage.QUIT)

            elif event.type == pygame.ACTIVEEVENT:
                rnd.state.options['needs_redraw'] = True

            elif rnd.state.stage == Stage.MENU:
                self.gui_menu.input(event, rnd)

            elif rnd.state.stage == Stage.CREDITS:
                self.gui_credits.input(event, rnd)

            elif rnd.state.stage == Stage.OPTIONS:
                self.gui_options.input(event, rnd)

            elif rnd.state.stage in [Stage.PLAYER_SHIPS, Stage.CPU_SHIPS]:
                self.gui_setupships.input(event, rnd)

            elif rnd.state.stage in [Stage.PLAYER_TURN, Stage.CPU_TURN]:
                self.gui_battleships.input(event, rnd)

            elif rnd.state.stage == Stage.NOTIFICATION:
                if event.type == pygame.MOUSEBUTTONDOWN or event.type == pygame.KEYDOWN:
                    rnd.state.options['skip'] = True

            elif rnd.state.stage == rnd.State.Stage.GAMEOVER:
                self.gui_gameover.input(event, rnd)

            if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
                rnd.state.switch(Stage.MENU)
